// Copyright 2004 Adam Megacz, see the COPYING file for licensing [GPL]
package org.ibex;
import org.xwt.translators.*;
import org.ibex.util.*;
import java.io.*;

// FEATURE: this could be cleaner
/** encapsulates a single font (a set of Glyphs) */
public class Font {

    private Font(Stream stream, int pointsize) { this.stream = stream; this.pointsize = pointsize; }

    private static boolean glyphRenderingTaskIsScheduled = false;

    public final int pointsize;                 ///< the size of the font
    public final Stream stream;                 ///< the stream from which this font was loaded
    public int max_ascent;                      ///< the maximum ascent, in pixels
    public int max_descent;                     ///< the maximum descent, in pixels
    boolean latinCharsPreloaded = false;        ///< true if a request to preload ASCII 32-127 has begun
    Glyph[] glyphs = new Glyph[65535];          ///< the glyphs that comprise this font

    public abstract static class Glyph {
        protected Glyph(Font font, char c) { this.font = font; this.c = c; }
        public final Font font;
        public final char c;
        public int baseline;                    ///< within the alphamask, this is the y-coordinate of the baseline
        public int advance;                     ///< amount to increment the x-coordinate
        public boolean isLoaded = false;        ///< true iff the glyph is loaded
        public int width = -1;                  ///< the width of the glyph
        public int height = -1;                 ///< the height of the glyph
        public byte[] data = null;              ///< the alpha channel samples for this font
    }


    // Statics //////////////////////////////////////////////////////////////////////

    private static final Freetype freetype = new Freetype();
    static final Queue glyphsToBeRendered = new Queue(255);
    private static Cache fontCache = new Cache(100);
    public static Font getFont(Stream stream, int pointsize) {
        Font ret = (Font)fontCache.get(stream, new Integer(pointsize));
        if (ret == null) fontCache.put(stream, new Integer(pointsize), ret = new Font(stream, pointsize));
        return ret;
    }


    // Methods //////////////////////////////////////////////////////////////////////

    /**
     *  Rasterize the glyphs of <code>text</code>.
     *
     *  If all the glyphs of <code>text</code> are not yet loaded,
     *  spawn a Task to load them and then invoke callback.  If all
     *  the glyphs <i>are</i> loaded, rasterize them to the
     *  PixelBuffer (if non-null).
     *
     *  @returns <code>(width&lt;&lt;32)|height</code> if all glyphs are loaded; else -1
     */
    public long rasterizeGlyphs(final String text, PixelBuffer pb, int textcolor,
                                int x, int y, int cx1, int cy1, int cx2, int cy2,
                                final Scheduler.Task callback) {
        boolean encounteredUnrenderedGlyph = false;
        int width = 0, height = 0;
        for(int i=0; i<text.length(); i++) {
            final char c = text.charAt(i);
            Glyph g = glyphs[c];
            if (g == null) {
                g = Platform.createGlyph(this, c);
                glyphs[c] = g;
            }
            if (!g.isLoaded) {
                //Log.debug(Font.class, "rasterizeGlyphs encountered unrasterized glyph " + g.c + " of font " + this);
                //System.out.println("rasterizeGlyphs encountered unrasterized glyph " + g.c + " of font " + this);
                glyphsToBeRendered.prepend(g);              // even if it's already in the queue, boost its priority
                encounteredUnrenderedGlyph = true;
            } else if (!encounteredUnrenderedGlyph) {
                if (pb != null) pb.drawGlyph(g, x + width, y + g.font.max_ascent - g.baseline, cx1, cy1, cx2, cy2, textcolor);
                width += g.advance;
                height = java.lang.Math.max(height, max_ascent + max_descent);
            }
        }

        if (!encounteredUnrenderedGlyph) return ((((long)width) << 32) | (long)(height & 0xffffffffL));

        if (callback != null) Scheduler.add(new Scheduler.Task() {
                public void perform() throws Exception {
                    // FEATURE this isn't terribly efficient... perhaps the task should go on the last glyph?
                    for(int i=0; i<text.length(); i++) {
                        Glyph g = glyphs[text.charAt(i)];
                        if (g == null || !g.isLoaded) { Scheduler.add(this); return; }
                    }
                    callback.perform();
                }});

        // preload the Latin-1 charset with low priority (we'll probably want it)
        if (!latinCharsPreloaded) {
            for(int i=48; i<57; i++) glyphsToBeRendered.append(glyphs[i] = Platform.createGlyph(this, (char)i));
            for(int i=32; i<47; i++) glyphsToBeRendered.append(glyphs[i] = Platform.createGlyph(this, (char)i));
            for(int i=57; i<128; i++) glyphsToBeRendered.append(glyphs[i] = Platform.createGlyph(this, (char)i));
            latinCharsPreloaded = true;
        }
        if (!glyphRenderingTaskIsScheduled) {
            Scheduler.add(glyphRenderingTask);
            glyphRenderingTaskIsScheduled = true;
        }
        return -1;
    }

    // FEATURE do we really need to be caching sizes?
    private static Cache sizeCache = new Cache(1000);
    public int textwidth(String s) { return (int)((textsize(s) >>> 32) & 0xffffffff); }
    public int textheight(String s) { return (int)(textsize(s) & 0xffffffffL); }
    public long textsize(String s) {
        Long l = (Long)sizeCache.get(s);
        if (l != null) return ((Long)l).longValue();
        long ret = rasterizeGlyphs(s, null, 0, 0, 0, 0, 0, 0, 0, null);
        if (ret != -1) sizeCache.put(s, new Long(ret));
        return ret == -1 ? 0 : ret;
    }

    static final Scheduler.Task glyphRenderingTask = new Scheduler.Task() { public void perform() {
        Glyph g = (Glyph)glyphsToBeRendered.remove(false);
        if (g == null) { glyphRenderingTaskIsScheduled = false; return; }
        Log.debug(Font.class, "glyphRenderingTask dequeued glyph " + g.c + " of font " + g.font);
        if (g.isLoaded) { perform(); /* tailcall to the next glyph */ return; }
        Log.debug(Glyph.class, "rendering glyph " + g.c);
        try { freetype.renderGlyph(g); } catch (IOException e) { Log.info(Freetype.class, e); }
        Scheduler.add(this);          // keep ourselves in the queue until there are no glyphs to render
        glyphRenderingTaskIsScheduled = true;
    } };
}
